from typing import Any, Self

from pydantic import BaseModel

from hiraconf.provider import Provider


class Parser:
    def __init__(self) -> None:
        self.data: dict[str, Any] = {}

    def merge(self, provider: Provider) -> Self:
        if self.data:
            self._merge(first=self.data, second=provider.load())
        else:
            self.data = provider.load()
        return self

    def _merge(self, first: dict[str, Any], second: dict[str, Any]) -> dict[str, Any]:
        for key in second:
            if key in first:
                if isinstance(first[key], dict) and isinstance(second[key], dict):
                    self._merge(first[key], second[key])
                elif first[key] != second[key]:
                    first[key] = second[key]
            else:
                first[key] = second[key]
        return first

    def join(self, provider: Provider) -> Self:
        if self.data:
            self._join(first=self.data, second=provider.load())
        else:
            self.data = provider.load()
        return self

    def _join(self, first: dict[str, Any], second: dict[str, Any]) -> dict[str, Any]:
        for key in second:
            if key in first:
                if isinstance(first[key], dict) and isinstance(second[key], dict):
                    self._join(first[key], second[key])
                elif first[key] != second[key]:
                    continue
            else:
                first[key] = second[key]
        return first

    def extract(self, config: type[BaseModel]) -> BaseModel:
        return config(**self.data)



import os
from pathlib import Path

from hiraconf.parser import Parser
from hiraconf.provider import EnvPrefixProvider, TomlFileProvider, YamlFileProvider

"""
These are all config files and Envvars that need to be merged together
"""

default_settings = Path(__file__).parent / "configs" / "default.yaml"
global_settings = Path(__file__).parent / "configs" / "global.yaml"
workspace_settings = Path(__file__).parent / "configs" / "workspace.toml"
os.environ["HIRACONF_ENV_VALUE"] = "100"

"""
This will parse each file and overwrite previous values and add new ones
"""

config = (
    Parser()
    .merge(YamlFileProvider(default_settings))
    .merge(YamlFileProvider(global_settings))
    .merge(TomlFileProvider(workspace_settings))
    .merge(EnvPrefixProvider("HIRACONF_"))
)

print(config.data)
# >>> {'settings': {'value': 3, 'default': True, 'global': True, 'workspace': True}, 'env_value': '100'}

"""
This will parse each file and NOT overwrite previous values, but add new ones
"""


config = (
    Parser()
    .join(YamlFileProvider(default_settings))
    .join(YamlFileProvider(global_settings))
    .join(TomlFileProvider(workspace_settings))
    .join(EnvPrefixProvider("HIRACONF_"))
)

print(config.data)
# >>> {'settings': {'value': 1, 'default': True, 'global': True, 'workspace': True}, 'env_value': '100'}



import os
from pathlib import Path

from hiraconf.provider import (
    EnvPrefixProvider,
    TomlFileProvider,
    TomlStringProvider,
    YamlFileProvider,
    YamlStringProvider,
)

testfiles = Path(__file__).parent / "testfiles"


def test_environment_prefix_provider() -> None:
    os.environ["HIRACONF_DEBUG"] = "True"
    os.environ["HIRACONF_VERSION"] = "1.0.0"

    loaded = EnvPrefixProvider("HIRACONF_").load()

    assert loaded == {"debug": "True", "version": "1.0.0"}


def test_toml_string_provider() -> None:
    toml_string = """[settings]
    debug = true
    version = "1.0.0"
    """
    loaded = TomlStringProvider(toml_string=toml_string).load()

    assert loaded == {"settings": {"debug": True, "version": "1.0.0"}}


def test_yaml_string_provider() -> None:
    yaml_string = """settings:
      debug: true
      version: 1.0.0
    """

    loaded = YamlStringProvider(yaml_string=yaml_string).load()

    assert loaded == {"settings": {"debug": True, "version": "1.0.0"}}


def test_toml_file_provider() -> None:
    loaded = TomlFileProvider(filepath=Path(testfiles, "second.toml")).load()

    assert loaded == {"settings": {"debug": True, "version": "1.1.0", "new": 10}}


def test_yaml_file_provider() -> None:
    loaded = YamlFileProvider(filepath=Path(testfiles, "first.yaml")).load()

    assert loaded == {"settings": {"debug": True, "version": "1.0.0"}}



import os

from hiraconf.parser import Parser
from hiraconf.provider import TomlStringProvider, YamlStringProvider

os.environ["HIRACONF_DEBUG"] = "True"
os.environ["HIRACONF_VERSION"] = "1.0.0"

yaml_string = """settings:
  debug: true
  version: 1.0.0
"""

toml_string = """[settings]
debug = true
version = "1.1.0"
new = 10
"""


def test_first_merge() -> None:
    data = Parser().merge(YamlStringProvider(yaml_string=yaml_string)).data

    assert data == {"settings": {"debug": True, "version": "1.0.0"}}


def test_second_merge() -> None:
    data = (
        Parser().merge(YamlStringProvider(yaml_string=yaml_string)).merge(TomlStringProvider(toml_string=toml_string))
    ).data

    assert data == {"settings": {"debug": True, "version": "1.1.0", "new": 10}}


def test_first_join() -> None:
    data = Parser().join(YamlStringProvider(yaml_string=yaml_string)).data

    assert data == {"settings": {"debug": True, "version": "1.0.0"}}


def test_second_join() -> None:
    data = (
        Parser().join(YamlStringProvider(yaml_string=yaml_string)).join(TomlStringProvider(toml_string=toml_string))
    ).data

    assert data == {"settings": {"debug": True, "version": "1.0.0", "new": 10}}



from pydantic import BaseModel

from hiraconf.parser import Parser
from hiraconf.provider import TomlStringProvider, YamlStringProvider


# Create the BaseModels for validation of configuration
class Settings(BaseModel):
    debug: bool
    version: str
    new: int | None = None


class Config(BaseModel):
    settings: Settings


# Define your config as file, env, string, toml, yaml or any provider that implements the Provider ABC

yaml_settings = """settings:
  debug: true
  version: 1.0.0
  """

parsed_yaml = Parser().merge(YamlStringProvider(yaml_string=yaml_settings)).extract(Config)

print(parsed_yaml)
# >>> settings=Settings(debug=False, version='1.0.0', new=None)

# If you have multiple configs and want to merge those, you can. Like overwriting the debug setting.

toml_settings = """[settings]
debug = false
"""

parsed_merged = (
    Parser()
    .merge(YamlStringProvider(yaml_string=yaml_settings))
    .merge(TomlStringProvider(toml_string=toml_settings))
    .extract(Config)
)

print(parsed_merged)
# >>> settings=Settings(debug=False, version='1.0.0', new=None)

# Or you join it and keep the original value if it exists as is

toml_join_settings = """[settings]
debug = false
new = 10
"""

parsed_joined = (
    Parser()
    .merge(YamlStringProvider(yaml_string=yaml_settings))
    .join(TomlStringProvider(toml_string=toml_join_settings))
    .extract(Config)
)

print(parsed_joined)
# >>> settings=Settings(debug=True, version='1.0.0', new=10)



import os
import tomllib
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

import yaml


class Provider(ABC):
    @abstractmethod
    def load(self) -> dict[str, Any]: ...


class TomlFileProvider(Provider):
    def __init__(self, filepath: Path) -> None:
        self.filepath = filepath

    def load(self) -> dict[str, Any]:
        if self.filepath.exists():
            with self.filepath.open("rb") as f:
                data = tomllib.load(f)
            return data
        raise FileNotFoundError(f"File does not exist: {self.filepath}")


class TomlStringProvider(Provider):
    def __init__(self, toml_string: str) -> None:
        self.toml_string = toml_string

    def load(self) -> dict[str, Any]:
        return tomllib.loads(self.toml_string)


class YamlFileProvider(Provider):
    def __init__(self, filepath: Path) -> None:
        self.filepath = filepath

    def load(self) -> dict[str, Any]:
        if self.filepath.exists():
            with self.filepath.open("r") as f:
                data = yaml.safe_load(f)
            return data
        raise FileNotFoundError(f"File does not exist: {self.filepath}")


class YamlStringProvider(Provider):
    def __init__(self, yaml_string: str) -> None:
        self.yaml_string = yaml_string

    def load(self) -> dict[str, Any]:
        return yaml.safe_load(self.yaml_string)


class EnvPrefixProvider(Provider):
    def __init__(self, prefix: str) -> None:
        self.prefix = prefix

    def load(self) -> dict[str, Any]:
        tmp: dict[str, Any] = {}
        for key, value in os.environ.items():
            if key.startswith(self.prefix.upper()):
                tmp[key.removeprefix(self.prefix).lower()] = value
        return tmp
